#!/usr/bin/env vpython3
import sys
import time
import random
import io
import datetime
import hashlib
import codecs
import struct
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12
import pdfrw as pdf
from pdfrw import pdfwriter as pdfw
from endesive import signer


class PdfBasicObject(pdf.PdfObject):
    def __init__(self, data):
        self.data = data


class PdfHexBytes(pdf.PdfObject):
    def __init__(self, data):
        self.data = data

    def __str__(self):
        return "<" + self.data.decode("utf-8") + ">"


class PdfNumber(PdfBasicObject):
    Format = "%d"

    def __str__(self):
        return self.Format % self.data


class PdfNumberB(PdfNumber):
    Format = "%08d"


class PdfNumberFloat(PdfNumber):
    Format = "%.5f"

    def __str__(self):
        s = (self.Format % self.data).rstrip("0").rstrip(".")
        return s


class PdfIndirect(tuple):
    def __str__(self):
        return "%d %d R" % self


class Signer(object):
    def __init__(self, fname, password):
        self.fname = fname
        self.password = password if password != "" else None
        self.compress = False
        self.objects = []
        with open(fname, "rb") as fi:
            self.datau = fi.read()
        self.startdata = len(self.datau)
        self.annotbutton = False
        s = b"startxref"
        i = self.datau.rfind(s)
        assert i != -1
        i += len(s)
        while self.datau[i] not in b"0123456789":
            i += 1
        j = i
        while self.datau[j] in b"0123456789":
            j += 1
        s = self.datau[i:j].decode()
        startprev = int(s, 10)
        self.startprev = startprev
        self.prev = pdf.PdfFileReader(
            fdata=self.datau, decrypt=(password is not None), password=password
        )

    def format_array(self, myarray):
        subarray = []
        bigarray = []
        count = 1000000
        for x in myarray:
            x = self.format(x)
            lenx = len(x) + 1
            count += lenx
            if count > 71:
                subarray = []
                bigarray.append(subarray)
                count = lenx
            subarray.append(x)
        return "[%s]" % "\n".join([" ".join(x) for x in bigarray])

    def format_dict(self, obj):
        if self.compress and obj.stream:
            pdfw.do_compress([obj])
        myarray = []
        for key, value in obj.iteritems():
            key = getattr(key, "encoded", None) or key
            key = self.format_obj(key)
            value = self.format(value)
            myarray.append(key + " " + value)
        result = "<<%s>>" % "\n".join(myarray)
        if obj.stream is not None:
            result = "%s\nstream\n%s\nendstream" % (result, obj.stream)
        return result

    def format_obj(self, obj):
        return str(obj)

    def format(self, obj):
        if isinstance(obj, pdf.PdfArray):
            return self.format_array(obj)
        elif isinstance(obj, pdf.PdfDict):
            return self.format_dict(obj)
        elif hasattr(obj, "indirect"):
            obj = getattr(obj, "encoded", None) or obj
        return self.format_obj(obj)

    def write(self, stream):
        positions = {}

        for i, x in enumerate(self.objects):
            if x is None:
                positions[i + 1] = 0
                continue
            positions[i + 1] = self.startdata + stream.tell()
            x = self.format(x)
            objstr = "%s 0 obj\n%s\nendobj\n" % (i + 1, x)
            stream.write(pdfw.convert_store(objstr))

        # xref table
        xref_location = self.startdata + stream.tell()
        stream.write(b"xref\n")
        stream.write(b"0 1\n")
        stream.write(b"0000000000 65535 f \n")
        keys = sorted(positions.keys())
        i = 0
        while i < len(keys):
            off = positions[keys[i]]
            if off == 0:
                while i < len(keys) and positions[keys[i]] == 0:
                    i += 1
                start = i
                while i < len(keys) and positions[keys[i]] != 0:
                    i += 1
                stream.write(b"%d %d\n" % (keys[start], i - start))
                i = start
                off = positions[keys[i]]
            stream.write(b"%010d %05d n \n" % (off, 0))
            i += 1

        # trailer
        stream.write(b"trailer\n")
        objstr = self.format(self.trailer)
        stream.write(pdfw.convert_store(objstr))
        # eof
        stream.write(b"\nstartxref\n%d\n%%%%EOF\n" % xref_location)

    def addObject(self, obj):
        self.objects.append(obj)
        return PdfIndirect((len(self.objects), 0))

    def copyarray(self, obj):
        dct = pdf.PdfArray()
        for v in obj:
            if v.indirect:
                v = PdfIndirect(v.indirect)
            elif isinstance(v, pdf.PdfArray):
                v = self.copydict(v)
            elif isinstance(v, pdf.PdfDict):
                v = self.copydict(v)
            dct.append(v)
        return dct

    def copydict(self, obj):
        dct = pdf.PdfDict()
        for k, v in obj.iteritems():
            if v.indirect:
                v = PdfIndirect(v.indirect)
            elif isinstance(v, pdf.PdfArray):
                v = self.copyarray(v)
            elif isinstance(v, pdf.PdfDict):
                v = self.copydict(v)
            dct[k] = v
        return dct

    def extend(self, obj):
        dct = pdf.PdfDict()
        dct.stream = obj.stream
        for k, v in obj.iteritems():
            if isinstance(v, pdf.PdfDict):
                if not v.indirect:
                    v = self.extend(v)
                    v = self.addObject(v)
                else:
                    v = self.extend(v)
            elif isinstance(v, list):
                v = pdf.PdfArray(v)
            dct[k] = v
        return dct

    def makepdf(self, zeros):
        root = self.prev.Root
        size = int(self.prev.Size, 10)

        while len(self.objects) < size - 1:
            self.objects.append(None)

        page0 = self.copydict(root.Pages.Kids[0])
        page0ref = PdfIndirect(root.Pages.Kids[0].indirect)

        obj10 = pdf.PdfDict()
        obj10ref = self.addObject(obj10)
        obj11 = pdf.PdfDict()
        obj11ref = self.addObject(obj11)
        obj12 = pdf.PdfDict()
        obj12ref = self.addObject(obj12)
        obj13 = pdf.PdfDict()
        obj13ref = self.addObject(obj13)
        obj14 = pdf.PdfDict()
        obj14ref = self.addObject(obj14)

        obj10.update(
            {
                pdf.PdfName("Type"): pdf.PdfName("TransformParams"),
                pdf.PdfName("P"): PdfNumber(2),
                pdf.PdfName("V"): pdf.PdfName("1.2"),
            }
        )
        obj11.update(
            {
                pdf.PdfName("Type"): pdf.PdfName("SigRef"),
                pdf.PdfName("TransformMethod"): pdf.PdfName("DocMDP"),
                pdf.PdfName("DigestMethod"): pdf.PdfName("SHA1"),
                pdf.PdfName("TransformParams"): obj10ref,
            }
        )
        obj12.update(
            {
                pdf.PdfName("Type"): pdf.PdfName("Sig"),
                pdf.PdfName("Filter"): pdf.PdfName("Adobe.PPKLite"),
                pdf.PdfName("SubFilter"): pdf.PdfName("adbe.pkcs7.detached"),
                pdf.PdfName("Name"): pdf.PdfString.from_unicode("Example User"),
                pdf.PdfName("Location"): pdf.PdfString.from_unicode("Los Angeles, CA"),
                pdf.PdfName("Reason"): pdf.PdfString.from_unicode("Testing"),
                pdf.PdfName("M"): pdf.PdfString.from_unicode("D:20200317214832+01'00'"),
                pdf.PdfName("Reference"): pdf.PdfArray([obj11ref]),
                pdf.PdfName("Contents"): PdfHexBytes(zeros),
                pdf.PdfName("ByteRange"): pdf.PdfArray(
                    [PdfNumberB(0), PdfNumberB(0), PdfNumberB(0), PdfNumberB(0)]
                ),
            }
        )
        obj13.update(
            {
                pdf.PdfName("FT"): pdf.PdfName("Sig"),
                pdf.PdfName("Type"): pdf.PdfName("Annot"),
                pdf.PdfName("Subtype"): pdf.PdfName("Widget"),
                pdf.PdfName("F"): PdfNumber(132),
                pdf.PdfName("T"): pdf.PdfString.from_unicode("Signature1"),
                pdf.PdfName("V"): obj12ref,
                pdf.PdfName("P"): page0ref,
                pdf.PdfName("Rect"): pdf.PdfArray(
                    [
                        PdfNumberFloat(0.0),
                        PdfNumberFloat(0.0),
                        PdfNumberFloat(0.0),
                        PdfNumberFloat(0.0),
                    ]
                ),
            }
        )
        obj14.update({pdf.PdfName("DocMDP"): obj12ref})
        obj15 = pdf.PdfDict()
        obj15.update(
            {
                pdf.PdfName("Fields"): pdf.PdfArray([obj13ref]),
                pdf.PdfName("SigFlags"): PdfNumber(3),
            }
        )
        obj15ref = self.addObject(obj15)
        if self.annotbutton:
            from pdf_annotate.annotations.image import Image
            from pdf_annotate.annotations.text import FreeText
            from pdf_annotate.config.appearance import Appearance
            from pdf_annotate.config.location import Location
            from pdf_annotate.util.geometry import identity

            annotation = "User signature text"
            x1, y1, x2, y2 = (470, 0, 570, 100)
            annotation = FreeText(
                Location(x1=x1, y1=y1, x2=x2, y2=y2, page=0),
                Appearance(
                    fill=[0, 0, 0],
                    stroke_width=1,
                    wrap_text=True,
                    font_size=12,
                    content=annotation,
                ),
            )
            pdfa = annotation.as_pdf_object(identity(), page=None)
            objapn = self.extend(pdfa[pdf.PdfName("AP")][pdf.PdfName("N")])
            objapnref = self.addObject(objapn)
            for name in (
                "BS",
                "C",
                "Contents",
                "DA",
                "Rect",
                # "Subtype",
            ):
                key = pdf.PdfName(name)
                v = pdfa[key]
                if isinstance(v, str):
                    v = v.replace("/", "//")
                    v = pdf.PdfString.from_unicode(v)
                elif isinstance(v, list):
                    v = pdf.PdfArray(v)
                obj13.update({key: v})

            objap = pdf.PdfDict()
            objap.update({pdf.PdfName("N"): objapnref})
            obj13.update(
                {
                    pdf.PdfName("SM"): pdf.PdfString.from_unicode("TabletPOSinline"),
                    pdf.PdfName("AP"): objap,
                }
            )
            self.objects[root.Pages.Kids[0].indirect[0] - 1] = page0
            annots = pdf.PdfArray([obj13ref])
            # if False and pdf.PdfName("Annots") in page0:
            if pdf.PdfName("Annots") in page0:
                page0annots = page0[pdf.PdfName("Annots")]
                if isinstance(page0annots, PdfIndirect):
                    annots.insert(0, page0annots)
                elif isinstance(page0annots, pdf.PdfArray):
                    annots = page0annots
                    annots.append(obj13ref)
            page0.update({pdf.PdfName("Annots"): annots})

        croot = self.copydict(root)
        croot.update(
            {pdf.PdfName("Perms"): obj14ref, pdf.PdfName("AcroForm"): obj15ref}
        )
        self.objects[root.indirect[0] - 1] = croot
        try:
            ID = self.prev.ID[0]
        except:
            b = hashlib.md5(self.datau).digest()
            ID = pdf.PdfString.from_bytes(b, bytes_encoding="hex")
        b = repr(random.random()).encode()
        b = hashlib.md5(b).digest()
        self.trailer = pdf.PdfDict(
            Size=len(self.objects),
            Root=PdfIndirect(root.indirect),
            Info=PdfIndirect(self.prev.Info.indirect),
            Prev=self.startprev,
            ID=pdf.PdfArray([ID, pdf.PdfString.from_bytes(b, bytes_encoding="hex")]),
        )
        if self.prev.private.pdfdict.encrypt:
            self.trailer.Encrypt = PdfIndirect(
                self.prev.private.pdfdict.encrypt.indirect
            )

    def sign(self, md, algomd):
        tspurl = "http://public-qlts.certum.pl/qts-17"
        tspurl = None
        fname = "demo2_user1.p12"
        with open(fname, "rb") as fp:
            p12 = pkcs12.load_key_and_certificates(
                fp.read(), b"1234", backends.default_backend()
            )
        contents = signer.sign(
            None, p12[0], p12[1], p12[2], algomd, True, md, None, False, tspurl
        )
        return contents

    def main(self):
        algomd = "sha1"
        aligned = False

        if aligned:
            zeros = b"0" * 0x4000
        else:
            md = getattr(hashlib, algomd)().digest()
            zeros = self.sign(md, algomd)
            zeros = zeros.hex().encode()

        self.makepdf(zeros)

        fo = io.BytesIO()
        self.write(fo)
        datas = fo.getvalue()

        br = [0, 0, 0, 0]
        bfrom = (" ".join([PdfNumberB.Format] * 4)) % tuple(br)
        bfrom = ("[" + bfrom + "]").encode()

        pdfbr1 = datas.find(zeros)
        pdfbr2 = pdfbr1 + len(zeros)
        br = [
            0,
            self.startdata + pdfbr1 - 1,
            self.startdata + pdfbr2 + 1,
            len(datas) - pdfbr2 - 1,
        ]
        bto = b"[%d %d %d %d]" % tuple(br)
        bto += b" " * (len(bfrom) - len(bto))
        assert len(bfrom) == len(bto)
        datas = datas.replace(bfrom, bto, 1)

        md = getattr(hashlib, algomd)()
        md.update(self.datau)
        b1 = datas[: br[1] - self.startdata]
        b2 = datas[br[2] - self.startdata :]
        md.update(b1)
        md.update(b2)
        md = md.digest()

        contents = self.sign(md, algomd)
        contents = contents.hex().encode("utf-8")
        if aligned:
            nb = len(zeros) - len(contents)
            contents += b"0" * nb
        datas = datas.replace(zeros, contents, 1)

        fname = self.fname.replace(".pdf", "-signed-pdfrw.pdf")
        with open(fname, "wb") as fp:
            fp.write(self.datau)
            fp.write(datas)


def main():
    if len(sys.argv) > 2:
        cls = Signer(sys.argv[1], sys.argv[2])
        cls.main()
    else:
        for fname, password in (
            ("pdf.pdf", ""),
            # ("pdf-encrypted.pdf", "1234"),
            # ("pdf-buggy.pdf", ""),
        ):
            cls = Signer(fname, password)
            cls.main()


main()



# -*- coding: utf-8 -*-
"""
    PdfAnnotator
    ~~~~~~~~~~~~
    The core annotator class.

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
import warnings

#from pdfrw import PdfReader
#from pdfrw import PdfWriter

from .annotations.image import Image
from .annotations.points import Ink
from .annotations.points import Line
from .annotations.points import Polygon
from .annotations.points import Polyline
from .annotations.rect import Circle
from .annotations.rect import Square
from .annotations.text import FreeText
from .config.metadata import Metadata
from .config.metadata import UNSET
from .graphics import ContentStream
from .util.geometry import identity
from .util.geometry import matrix_multiply
from .util.geometry import normalize_rotation
from .util.geometry import rotate
from .util.geometry import scale
from .util.geometry import translate
from .util.validation import NUMERIC_TYPES


NAME_TO_ANNOTATION = {
    'square': Square,
    'circle': Circle,
    'line': Line,
    'polygon': Polygon,
    'polyline': Polyline,
    'ink': Ink,
    'text': FreeText,
    'image': Image,
}


class PDF(object):

    def __init__(self, pdf_reader):
        self._reader = pdf_reader
        self.pdf_version = self._reader.private.pdfdict.version

    def get_page(self, page_number):
        if page_number > len(self._reader.pages) - 1:
            raise ValueError('Page number {} out of bounds ({} pages)'.format(
                page_number,
                len(self._reader.pages),
            ))
        return self._reader.pages[page_number]

    def get_rotation(self, page_number):
        """Returns the rotation of a specified page."""
        page = self.get_page(page_number)
        rotate = int(page.inheritable.Rotate or 0)
        return normalize_rotation(rotate)


class PdfAnnotator(object):

    def __init__(self, file_or_reader, scale=None, compress=True):
        """Draw annotations directly on PDFs. Annotations are always drawn on
        as if you're drawing them in a viewer, i.e. they take into account page
        rotation and weird, translated coordinate spaces.

        :param str|PdfReader file_or_reader: filename of PDF or pdfrw.PdfReader
        :param number|tuple|None scale: number by which to scale coordinates
            to get to default user space. Use this if, for example, your points
            in the coordinate space of the PDF viewed at a dpi. In this case,
            scale would be 72/dpi. Can also specify a 2-tuple of x and y scale.
        :param bool compress: whether to output flate-compressed PDFs
        """
        if isinstance(file_or_reader, str):
            file_or_reader = PdfReader(file_or_reader)
        self._pdf = PDF(file_or_reader)
        self._scale = self._expand_scale(scale)
        self._dimensions = {}
        self._compress = compress

    def _expand_scale(self, scale):
        if scale is None:
            return 1, 1
        elif isinstance(scale, NUMERIC_TYPES):
            return (scale, scale)
        return scale

    def set_page_dimensions(self, dimensions, page_number):
        """Set dimensions for a given page number. If set, the dimensions for
        this page override the document-wide rotation and scale settings.

        :param tuple|None dimensions: As a convenient alternative to scale and
            you can pass in the dimensions of your sheet when viewed in a
            certain setting. For example, an 8.5"x11" PDF, rotated at 90° and
            rastered at 150 dpi, would produce dimensions of (1650, 1275). If
            you pass this in, you can then specify your coordinates in this
            coordinate space.
        :param int page_number:
        """
        self._dimensions[page_number] = dimensions

    def get_page_bounding_box(self, page_number):
        page = self._pdf.get_page(page_number)
        # PDF bounding boxes are complicated. We choose to use the CropBox, if
        # it's available, because that's what Acrobat uses to display the
        # actual PDF, and what pdfinfo uses to determine the PDF's dimensions.
        # If CropBox isn't available, we use MediaBox. We ignore the TrimBox
        # because Acrobat also ignores this when displaying the PDF.
        crop_box = page.inheritable.CropBox
        if crop_box is not None:
            return [float(n) for n in crop_box]
        return [float(n) for n in page.inheritable.MediaBox]

    def get_size(self, page_number):
        """Returns the size of the specified page's bounding box (pts),
        accounting for page rotation.

        :param int page_number:
        :returns tuple: If page is rotated 90° or 270°, the returned value will
            be (height, width) in PDF user space. Otherwise the returned value
            will be (width, height).
        """
        x1, y1, x2, y2 = self.get_page_bounding_box(page_number)
        rotation = self._pdf.get_rotation(page_number)

        if rotation in (0, 180):
            return (x2 - x1, y2 - y1)

        return (y2 - y1, x2 - x1)

    def add_annotation(
        self,
        annotation_type,
        location,
        appearance,
        metadata=None,
    ):
        """Add an annotation of the given type, with the given parameters, to
        the given location of the PDF.

        :param str annotation_type: E.g. 'square'
        :param Location location: Annotation's Location object, specified in
            the coordinate system of the client. Coordinates will be
            transformed to PDF user space via get_transform.
        :param Appearance appearance:
        :param Metadata|None|UNSET metadata: Metadata object. If UNSET, no
            metadata is written on the entire annotation. If None, default
            metadata is used.
        """
        self._before_add(location)
        metadata = self._resolve_metadata(metadata)
        self._validate_appearance_stream(appearance)
        annotation = self.get_annotation(
            annotation_type,
            location,
            appearance,
            metadata,
        )
        self._add_annotation(annotation)

    @staticmethod
    def _resolve_metadata(metadata):
        if isinstance(metadata, Metadata):
            return metadata
        elif metadata is None:
            return Metadata()
        elif metadata is UNSET:
            return None
        else:
            raise ValueError('Invalid metadata')

    @staticmethod
    def _validate_appearance_stream(appearance):
        stream = appearance.appearance_stream
        if stream is not None and not isinstance(stream, ContentStream):
            raise ValueError(
                'Invalid appearance stream format: {}'.format(type(stream)))

    def _before_add(self, location):
        # Steps to take before trying to add an annotation to `location`
        page = self._pdf.get_page(location.page)
        user_unit = page.inheritable.UserUnit
        if user_unit not in (1, None):
            warnings.warn(
                'Unsupported UserUnit (value: {})'.format(user_unit)
            )

    def get_annotation(self, annotation_type, location, appearance, metadata):
        # TODO filter on valid PDF versions, by type
        annotation_cls = NAME_TO_ANNOTATION.get(annotation_type)
        if annotation_cls is None:
            raise ValueError('Invalid/unsupported annotation type: {}'.format(
                annotation_type
            ))

        annotation = annotation_cls(location, appearance, metadata)
        annotation.validate(self._pdf.pdf_version)
        return annotation

    def get_scale(self, page_number):
        """Public API to get the x and y scales of the given page.

        :param int page_number:
        :returns 2-tuple: (x_scale, y_scale)
        """
        rotation = self._pdf.get_rotation(page_number)
        bounding_box = self.get_page_bounding_box(page_number)
        return self._get_scale(page_number, bounding_box, rotation)

    def get_rotation(self, page_number):
        """Public API to get the rotation of the give page.

        :param int page_number:
        :returns int: integer where i % 90 == 0
        """
        return self._pdf.get_rotation(page_number)

    def _get_scale(self, page_number, bounding_box, rotation):
        W = bounding_box[2] - bounding_box[0]
        H = bounding_box[3] - bounding_box[1]

        dimensions = self._dimensions.get(page_number)
        if dimensions is not None:
            # User-specified dimensions for a particular page just give us the
            # scaling factor to use for that page.
            width_d, height_d = dimensions
            width_pts, height_pts = W, H
            if rotation in (90, 270):
                width_pts, height_pts = H, W
            x_scale = (width_pts / float(width_d))
            y_scale = (height_pts / float(height_d))
        else:
            x_scale, y_scale = self._scale

        return x_scale, y_scale

    def get_transform(self, page_number, rotation):
        bounding_box = self.get_page_bounding_box(page_number)
        _scale = self._get_scale(page_number, bounding_box, rotation)
        return self._get_transform(bounding_box, rotation, _scale)

    @staticmethod
    def _get_transform(bounding_box, rotation, _scale):
        """Get the transformation required to go from the user's desired
        coordinate space to PDF user space, taking into account rotation,
        scaling, translation (for things like weird media boxes).
        """
        # Unrotated width and height, in pts
        W = bounding_box[2] - bounding_box[0]
        H = bounding_box[3] - bounding_box[1]

        scale_matrix = scale(*_scale)

        x_translate = 0 + bounding_box[0]
        y_translate = 0 + bounding_box[1]
        mb_translate = translate(x_translate, y_translate)

        # Because of how rotation works the point isn't rotated around an axis,
        # but the axis itself shifts. So we have to represent the rotation as
        # rotation + translation.
        rotation_matrix = rotate(rotation)

        translate_matrix = identity()
        if rotation == 90:
            translate_matrix = translate(W, 0)
        elif rotation == 180:
            translate_matrix = translate(W, H)
        elif rotation == 270:
            translate_matrix = translate(0, H)

        # Order matters here - the transformation matrices are applied in
        # reverse order. So first we scale to get the points in PDF user space,
        # since all other operations are in that space. Then we rotate and
        # scale to capture page rotation, then finally we translate to account
        # for offset media boxes.
        transform = matrix_multiply(
            mb_translate,
            translate_matrix,
            rotation_matrix,
            scale_matrix,
        )
        return transform

    def _add_annotation(self, annotation):
        """Add the annotation to the PDF document, transforming annotation
        metadata and content stream to PDF user space.
        """
        page = self._pdf.get_page(annotation.page)
        transform = self.get_transform(
            annotation.page,
            self._pdf.get_rotation(annotation.page),
        )
        annotation_obj = annotation.as_pdf_object(transform, page)

        if page.Annots:
            page.Annots.append(annotation_obj)
        else:
            page.Annots = [annotation_obj]

    def write(self, filename=None, overwrite=False):
        if filename is None and not overwrite:
            raise ValueError(
                'Must specify either output filename or overwrite flag'
            )
        if overwrite:
            filename = self._filename

        writer = PdfWriter(
            version=self._pdf.pdf_version,
            compress=self._compress,
        )
        writer.write(fname=filename, trailer=self._pdf._reader)



#!/usr/bin/env vpython3
class Main:
    def __init__(self, fname):
        with open(fname, "rb") as fi:
            self.data = fi.read()
        self.xrefs = []

    def find(self, s, start=0):
        i = self.data.find(s, start)
        if i >= 0:
            i += len(s)
        return i

    def readline(self, start):
        i1 = i0 = start
        while self.data[i1] not in b'\r\n':
            i1 += 1
        n = i1
        while self.data[n] in b'\r\n':
            n += 1
        return self.data[i0:i1].strip(), n

    def xref(self):
        xref = 0
        while True:
            xref = self.find(b'\nxref\n', xref)
            if xref < 0:
                break
            n = xref
            while True:
                line, n = self.readline(n)
                line = line.split()
                try:
                    assert len(line) == 2
                    off, cnt = int(line[0]), int(line[1])
                except:
                    break
                for i in range(cnt):
                    line, n = self.readline(n)
                    line = line.split()
                    if line[-1] == b'n':
                        offset = int(line[0])
                        line1 = self.readline(offset)[0].split()
                        try:
                            assert len(line1) == 3 and int(line1[0]) == off + i
                        except:
                            print('bad xref:', line, line1, 'off:', off+i, len(line1))
                        print(line1, off+i, offset)
                        #self.xref.append(
    def byterange(self):
        start = self.find(b'/ByteRange')
        i0 = self.find(b'[', start)
        i1 = self.find(b']', start)
        line = self.data[i0:i1-1]
        line = line.split()
        br = [int(line[0]), int(line[1]), int(line[2]), int(line[3])]
        print(br, [hex(i) for i in br])
        c = self.data[br[1]]
        print('[{:06x}]: {:d} {:02x}'.format(br[1], c, c))
        c = self.data[br[2]]
        print('[{:06x}]: {:d} {:02x}'.format(br[2], c, c))
        print(len(self.data), br[2]+br[3])
        contents = self.data[br[0] + br[1] + 1 : br[2] - 1]
        print('[contents]:', len(contents))

    def main(self):
        self.xref()
        self.byterange()

for fname in (
    #"pdf-certum.pdf",
    #"pdf-acrobat.pdf",
    'pdf-signed-java.pdf',
    'pdf-signed-pypdf.pdf',
    'pdf-encrypted-signed-java.pdf',
    'pdf-encrypted-signed-pypdf.pdf',
):
    print('*'*20, fname)
    try:
        cls = Main(fname)
    except IOError:
        continue
    cls.main()



# -*- coding: utf-8 -*-
"""
    Geometry
    ~~~~~~~~
    Geometry utility functions

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
import math


def normalize_rotation(rotate):
    if rotate % 90:
        raise ValueError('Invalid Rotate value: {}'.format(rotate))
    while rotate < 0:
        rotate += 360
    while rotate >= 360:
        rotate -= 360
    return rotate


def to_radians(degrees):
    return degrees * math.pi / 180


def rotate(degrees):
    """Return a homogenous rotation matrix by degrees

    :params int degrees: integer degrees
    :returns list: matrix rotated by degrees, 6-item list
    """
    radians = to_radians(degrees)
    return [
        math.cos(radians),
        math.sin(radians),
        -math.sin(radians),
        math.cos(radians),
        0,
        0,
    ]


def translate(x, y):
    return [1, 0, 0, 1, x, y]


def scale(x_scale, y_scale):
    return [x_scale, 0, 0, y_scale, 0, 0]


def identity():
    return [1, 0, 0, 1, 0, 0]


def matrix_multiply(*args):
    """Multiply a series of matrices. len(args) must be at least two. If more
    than two matrices are specified, the multiplications are chained, with the
    left-most matrices being multiplied first.

    E.g. matrix_multiply(A, B, C) => (A*B)*C

    What this means for combining affine transformations is that they are
    applied in reverse order. For instance, to perform rotation R, scale S, and
    translation T, in that order, you would use:

    matrix_multiply(T, S, R)

    Each matrix is a 6-item homogenous matrix.
    """
    if len(args) < 2:
        raise ValueError('Cannot multiply fewer than two matrices')
    r = _matrix_multiply(args[0], args[1])
    for m in args[2:]:
        r = _matrix_multiply(r, m)
    return r


def matrix_inverse(matrix):
    """Invert a 6-item homogenous transform matrix.

    A transform matrix [a, b, c, d, e, f] is part of a 3x3 matrix:

        a b 0
        c d 0
        e f 1

    Invert it using the formula for the inverse of a 3x3 matrix from
    http://mathworld.wolfram.com/MatrixInverse.html
    """
    a, b, c, d, e, f = matrix

    determinant = float(a * d - b * c)
    unscaled = [d, -b, -c, a, c * f - d * e, b * e - a * f]

    return [x / determinant for x in unscaled]


def _matrix_multiply(A, B):
    a00, a01, a10, a11, a20, a21 = A

    b00, b01, b10, b11, b20, b21 = B
    b02, b12, b22 = 0, 0, 1

    # We don't have to compute all entries of the new matrix during
    # multiplication, since any multiple of affine transformations is an affine
    # transformation, and therefore homogenous coordinates.
    c00 = b00 * a00 + b01 * a10 + b02 * a20
    c01 = b00 * a01 + b01 * a11 + b02 * a21

    c10 = b10 * a00 + b11 * a10 + b12 * a20
    c11 = b10 * a01 + b11 * a11 + b12 * a21

    c20 = b20 * a00 + b21 * a10 + b22 * a20
    c21 = b20 * a01 + b21 * a11 + b22 * a21

    return [c00, c01, c10, c11, c20, c21]


def transform_point(point, matrix):
    """Transform point by matrix.

    :param list point: 2-item list
    :param list matrix: 6-item list representing transformation matrix
    :returns list: 2-item transformed point
    """
    x, y = point
    a, b, c, d, e, f = matrix
    # This leaves out some unnecessary stuff from the fact that the matrix is
    # homogenous coordinates.
    new_x = x * a + y * c + e
    new_y = x * b + y * d + f
    return [new_x, new_y]


def transform_vector(vector, matrix):
    """Transform a vector by a matrix. This is similar to transform_point,
    except that translation isn't honored. Think of a vector as displacement in
    space, and a point as, well, a point in space.

    :param list vector: 2-item list
    :param list matrix: 6-item list representing transformation matrix
    :returns list: 2-item transformed point
    """
    x, y = vector
    a, b, c, d, _, _ = matrix
    new_x = x * a + y * c
    new_y = x * b + y * d
    return [new_x, new_y]


def transform_rect(rect, matrix):
    """Transform a rectangle specified by two points (lower left and upper
    right) by a transformation matrix.

    :param list rect: [x1, y1, x2, y2]
    :param list matrix: transformation matrix
    :returns list: [x1, y1, x2, y2] with transformed points
    """
    x1, y1 = transform_point(rect[:2], matrix)
    x2, y2 = transform_point(rect[2:], matrix)

    x1, x2 = sorted([x1, x2])
    y1, y2 = sorted([y1, y2])

    return [x1, y1, x2, y2]



# -*- coding: utf-8 -*-
"""
    Appearance Config
    ~~~~~~~~~~~~~~~~~
    Configuration for an annotation's appearance.

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
import attr

from .constants import ALLOWED_ALIGNS
from .constants import ALLOWED_BASELINES
from .constants import ALLOWED_LINE_CAPS
from .constants import ALLOWED_LINE_JOINS
from .constants import BLACK
from .constants import DEFAULT_BORDER_STYLE
from .constants import DEFAULT_CONTENT
from .constants import DEFAULT_FONT_SIZE
from .constants import DEFAULT_LINE_SPACING
from .constants import DEFAULT_STROKE_WIDTH
from .constants import GRAPHICS_STATE_NAME
from .constants import TEXT_ALIGN_LEFT
from .constants import TEXT_BASELINE_MIDDLE
from .graphics_state import GraphicsState
from ..graphics import ContentStream
from ..graphics import FillColor
from ..graphics import GraphicsState as CSGraphicsState
from ..graphics import Stroke
from ..graphics import StrokeAndFill
from ..graphics import StrokeColor
from ..graphics import StrokeWidth
from ..util.validation import between
from ..util.validation import Boolean
from ..util.validation import Color
from ..util.validation import Enum
from ..util.validation import Field
from ..util.validation import Number
from ..util.validation import positive
from ..util.validation import String
from ..util.validation import validate_dash_array


def is_transparent(color):
    # E.g. a soothing gray: [0, 0, 0, 0.5]
    if color is None:
        return False
    return len(color) == 4 and color[-1] < 1


@attr.s
class Appearance(object):
    # Stroke attributes
    stroke_color = Color(default=BLACK)
    stroke_width = Number(default=DEFAULT_STROKE_WIDTH, validator=positive)
    border_style = String(default=DEFAULT_BORDER_STYLE)
    dash_array = Field(list, default=None, validator=validate_dash_array)
    line_cap = Enum(ALLOWED_LINE_CAPS, default=None)
    line_join = Enum(ALLOWED_LINE_JOINS, default=None)
    miter_limit = Number(default=None, validator=positive)
    stroke_transparency = Number(default=None, validator=between(0, 1))

    # Fill attributes
    fill = Color(default=None)
    fill_transparency = Number(default=None, validator=between(0, 1))

    # Text attributes
    content = String(default=DEFAULT_CONTENT)
    font_size = Number(default=DEFAULT_FONT_SIZE, validator=positive)
    text_align = Enum(ALLOWED_ALIGNS, default=TEXT_ALIGN_LEFT)
    text_baseline = Enum(ALLOWED_BASELINES, default=TEXT_BASELINE_MIDDLE)
    line_spacing = Number(default=DEFAULT_LINE_SPACING, validator=positive)
    wrap_text = Boolean(default=True)

    # Image attributes
    image = String(default=None)

    # Advanced attributes
    appearance_stream = Field(ContentStream, default=None)
    xobjects = Field(dict, default=None)
    graphics_states = Field(dict, default=None)
    fonts = Field(dict, default=None)

    def copy(self, **kwargs):
        A = Appearance(**kwargs)
        for k, v in self.__dict__.items():
            if k not in kwargs:
                setattr(A, k, v)
        return A

    def _get_stroke_transparency(self):
        stroke_transparency = None
        if is_transparent(self.stroke_color):
            stroke_transparency = self.stroke_color[-1]
        if self.stroke_transparency is not None:
            stroke_transparency = self.stroke_transparency
        return stroke_transparency

    def _get_fill_transparency(self):
        fill_transparency = None
        if is_transparent(self.fill):
            fill_transparency = self.fill[-1]
        if self.fill_transparency is not None:
            fill_transparency = self.fill_transparency
        return fill_transparency

    def get_graphics_state(self):
        """Return a GraphicsState config from the appearance's graphics-state-
        applicable params.

        :returns GraphicsState:
        """
        return GraphicsState(
            dash_array=self.dash_array,
            line_cap=self.line_cap,
            line_join=self.line_join,
            miter_limit=self.miter_limit,
            stroke_transparency=self._get_stroke_transparency(),
            fill_transparency=self._get_fill_transparency(),
        )


def set_appearance_state(stream, A):
    """Update the graphics command stream to reflect appearance properties.

    :param ContentStream stream: current content stream
    :param Appearance A: appearance object
    """
    # Add in the `gs` command, which will execute the named graphics state from
    # the Resources dict, and set CA and/or ca values. The annotations
    # themselves will need to ensure that the proper ExtGState object is
    # present in the Resources dict.
    graphics_state = A.get_graphics_state()
    if graphics_state.has_content():
        stream.add(CSGraphicsState(GRAPHICS_STATE_NAME))

    stream.extend([
        StrokeColor(*A.stroke_color[:3]),
        StrokeWidth(A.stroke_width),
    ])

    # TODO support more color spaces - CMYK and GrayScale
    if A.fill is not None:
        stream.add(FillColor(*A.fill[:3]))


def stroke_or_fill(stream, A):
    if A.fill is not None:
        stream.add(StrokeAndFill())
    else:
        stream.add(Stroke())



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12
from endesive import email


def main():
    with open('ca/demo2_user1.p12', 'rb') as fp:
        p12 = pkcs12.load_key_and_certificates(fp.read(), b'1234', backends.default_backend())
    datau = open('smime-unsigned.txt', 'rb').read()
    datas = email.sign(datau,
        p12[0], p12[1], p12[2],
        'sha256',
        attrs=True
    )
    open('smime-signed-attr.txt', 'wb').write(datas)


main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='endesive',
    version=__import__('endesive').__version__,
    description='Library for digital signing and verification of digital signatures in mail, PDF and XML documents.',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/m32/endesive',
    author='Grzegorz Makarewicz',
    author_email='mak@trisoft.com.pl',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Communications :: Email',
        'Topic :: Security :: Cryptography',
        'Topic :: Office/Business',
        'Topic :: Text Processing',
        'Topic :: Multimedia :: Graphics',
    ],
    keywords='cryptography pki x509 smime email pdf pkcs11 asn1 xades',
    packages=find_packages(where=".", exclude=['examples', 'tests']),
    package_dir={"": "."},
    package_data={"": ["*.ttf"]},
    platforms=["all"],
    install_requires=['certvalidator', 'cryptography', 'asn1crypto', 'oscrypto', 'lxml', 'pykcs11', 'Pillow', 'pytz', 'requests','paramiko', 'attrs'],
    test_suite="tests",
)



"""
    Graphics
    ~~~~~~~~~~~~
    Classes for manipulating PDF graphics streams.

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
from __future__ import division

from collections import namedtuple
from functools import total_ordering

from .util.geometry import matrix_multiply
from .util.geometry import transform_point
from .util.geometry import transform_vector

ZERO_TOLERANCE = 0.00000000000001


class ContentStream(object):
    """An abstraction over a PDF content stream, e.g. (draw a green rectangle):
    '0 1 0 rg 3 w 1 1 2 2 re S'.

    This abstraction will allow users to draw more precise annotations, while
    still allowing this library to transform those annotations to be properly
    drawn in PDF user space. For instance, a user could do:

    content_stream = ContentStream([
        StrokeColor(1, 0, 0),
        StrokeWidth(5),
        Move(10, 10),
        Line(20, 20),
        Stroke(),
    ])
    annotator.add_annotation(
        'square',
        location=Location(10, 10, 20, 20, page=0),
        appearance=Appearance(appearance_stream=content_stream),
    )

    This would draw a "square" annotation as a line, which is kind of silly,
    but it shows the flexibility for the user to draw more complex annotations.
    Behind the scenes, the pdf-annotator library transforms the Move and Line
    operations to be properly placed in PDF user space.
    """

    def __init__(self, commands=None):
        self.commands = commands or []

    def __eq__(self, other):
        if not isinstance(other, ContentStream):
            return False

        return self.resolve() == other.resolve()

    def add(self, command):
        self.commands.append(command)

    def extend(self, commands):
        self.commands.extend(commands)

    def transform(self, transform):
        return ContentStream([
            command.transform(transform) for command in self.commands
        ])

    def resolve(self):
        return ' '.join(
            command.resolve() for command in self.commands
        )

    @staticmethod
    def join(stream1, stream2):
        """Combine two content streams."""
        return ContentStream(stream1.commands + stream2.commands)


@total_ordering
class BaseCommand(object):
    COMMAND = ''
    NUM_ARGS = 0

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False

        return self.resolve() == other.resolve()

    def __ne__(self, other):
        # yes you really have to define this
        return not self.__eq__(other)

    def __lt__(self, other):
        raise TypeError(
            'Comparison not supported between instances of {} and {}'.format(
                self.__class__.__name__,
                other.__class__.__name__,
            ),
        )

    def transform(self, t):
        return self

    def resolve(self):
        return self.COMMAND

    @classmethod
    def _get_tokens(cls, idx, tokens):
        return tokens[idx-cls.NUM_ARGS:idx]

    @classmethod
    def from_tokens(cls, idx, tokens):
        return cls(*cls._get_tokens(idx, tokens))


class TupleCommand(type):
    def __new__(cls, name, parents, attrs):
        namedtuple_klass = namedtuple(name + '_namedtuple', attrs['ARGS'])

        if 'NUM_ARGS' not in attrs:
            attrs['NUM_ARGS'] = len(attrs['ARGS'])

        attrs.pop('ARGS')

        # don't put object at beginning of MRO
        if parents == (object,):
                parents = ()

        new_parents = (*parents, BaseCommand, namedtuple_klass)

        return super(TupleCommand, cls).__new__(cls, name, new_parents, attrs)

    def __init__(cls, name, parents, attrs):
        if cls.resolve is BaseCommand.resolve:
            def resolve(self):
                return ' '.join([*self] + [self.COMMAND])

            setattr(cls, 'resolve', resolve)


class FloatTupleCommand(TupleCommand):
    def __init__(cls, name, parents, attrs):
        if cls.resolve is BaseCommand.resolve:
            def resolve(self):
                return ' '.join([format_number(n) for n in self] + [self.COMMAND])

            setattr(cls, 'resolve', resolve)

        @classmethod
        def from_tokens(cls, idx, tokens):
            return cls(*map(float, cls._get_tokens(idx, tokens)))

        setattr(cls, 'from_tokens', from_tokens)


class StrokeColor(metaclass=FloatTupleCommand):
    COMMAND = 'RG'
    ARGS = ['r', 'g', 'b']


class StrokeWidth(metaclass=FloatTupleCommand):
    COMMAND = 'w'
    ARGS = ['width']


class FillColor(metaclass=FloatTupleCommand):
    COMMAND = 'rg'
    ARGS = ['r', 'g', 'b']


class BeginText(BaseCommand):
    COMMAND = 'BT'


class EndText(BaseCommand):
    COMMAND = 'ET'


class Stroke(BaseCommand):
    COMMAND = 'S'


class CloseAndStroke(BaseCommand):
    COMMAND = 's'


class StrokeAndFill(BaseCommand):
    COMMAND = 'B'


class StrokeAndFillEvenOdd(BaseCommand):
    COMMAND = 'B*'


class Fill(BaseCommand):
    COMMAND = 'f'


class ReadOnlyFill(BaseCommand):
    # PDF reading should accept F as equivalent to f, but only write f.
    # - Table 60 in PDF spec.
    COMMAND = 'F'

    def resolve(self):
        return Fill.COMMAND


class FillEvenOdd(BaseCommand):
    COMMAND = 'f*'


class CloseFillAndStroke(BaseCommand):
    # equivalent to h B
    COMMAND = 'b'


class CloseFillAndStrokeEvenOdd(BaseCommand):
    # equivalent to h B*
    COMMAND = 'b*'


class EndPath(BaseCommand):
    # End path without filling or stroking; used for clipping paths.
    COMMAND = 'n'


class Save(BaseCommand):
    COMMAND = 'q'


class Restore(BaseCommand):
    COMMAND = 'Q'


class Close(BaseCommand):
    COMMAND = 'h'


class Font(metaclass=TupleCommand):
    COMMAND = 'Tf'
    ARGS = ['font', 'font_size']

    def resolve(self):
        return '/{} {} {}'.format(
            self.font,
            format_number(self.font_size),
            self.COMMAND,
        )

    @classmethod
    def from_tokens(cls, idx, tokens):
        # PDF spec calls font_size a "scale parameter" which implies > 0, but it
        # doesn't declare constraints on it. Unclear if/how we should validate.
        font, font_size = cls._get_tokens(idx, tokens)
        return cls(font, float(font_size))


class Text(metaclass=TupleCommand):
    COMMAND = 'Tj'
    ARGS = ['text']

    def resolve(self):
        return '({}) {}'.format(self.text, self.COMMAND)


class XObject(metaclass=TupleCommand):
    COMMAND = 'Do'
    ARGS = ['name']

    def resolve(self):
        return '/{} {}'.format(self.name, self.COMMAND)


class GraphicsState(metaclass=TupleCommand):
    COMMAND = 'gs'
    ARGS = ['name']

    def resolve(self):
        return '/{} {}'.format(self.name, self.COMMAND)


class Rect(metaclass=FloatTupleCommand):
    COMMAND = 're'
    ARGS = ['x', 'y', 'width', 'height']

    def transform(self, t):
        x, y = transform_point((self.x, self.y), t)
        width, height = transform_vector((self.width, self.height), t)
        return Rect(x, y, width, height)


class Move(metaclass=FloatTupleCommand):
    COMMAND = 'm'
    ARGS = ['x', 'y']

    def transform(self, t):
        x, y = transform_point((self.x, self.y), t)
        return Move(x, y)


class Line(metaclass=FloatTupleCommand):
    COMMAND = 'l'
    ARGS = ['x', 'y']

    def transform(self, t):
        x, y = transform_point((self.x, self.y), t)
        return Line(x, y)


class Bezier(metaclass=FloatTupleCommand):
    """Cubic bezier curve, from the current point to (x3, y3), using (x1, y1)
    and (x2, y2) as control points.
    """
    COMMAND = 'c'
    ARGS = ['x1', 'y1', 'x2', 'y2', 'x3', 'y3']

    def transform(self, t):
        x1, y1 = transform_point((self.x1, self.y1), t)
        x2, y2 = transform_point((self.x2, self.y2), t)
        x3, y3 = transform_point((self.x3, self.y3), t)
        return Bezier(x1, y1, x2, y2, x3, y3)


class BezierV(metaclass=FloatTupleCommand):
    """Cubic bezier curve, from the current point to (x3, y3), using (x2, y2)
    and (x3, y3) as control points.
    """
    COMMAND = 'v'
    ARGS = ['x2', 'y2', 'x3', 'y3']

    def transform(self, t):
        x2, y2 = transform_point((self.x2, self.y2), t)
        x3, y3 = transform_point((self.x3, self.y3), t)
        return BezierV(x2, y2, x3, y3)


class BezierY(metaclass=FloatTupleCommand):
    """Cubic bezier curve, from the current point to (x3, y3), using (x1, y1)
    and (x3, y3) as control points.
    """
    COMMAND = 'y'
    ARGS = ['x1', 'y1', 'x3', 'y3']

    def transform(self, t):
        x1, y1 = transform_point((self.x1, self.y1), t)
        x3, y3 = transform_point((self.x3, self.y3), t)
        return BezierY(x1, y1, x3, y3)


class MatrixCommand(TupleCommand):
    def __new__(cls, name, parents, attrs):
        attrs['ARGS'] = ['matrix']
        attrs['NUM_ARGS'] = 6

        return super(MatrixCommand, cls).__new__(cls, name, parents, attrs)

    def __init__(cls, name, parents, attrs):
        def transform(self, t):
            return cls(matrix_multiply(t, self.matrix))

        def resolve(self):
            return ' '.join([format_number(n) for n in self.matrix] + [self.COMMAND])

        @classmethod
        def from_tokens(cls, idx, tokens):
            return cls([float(tok) for tok in cls._get_tokens(idx, tokens)])

        def __init__(self, matrix):
            if len(matrix) != 6:
                raise ValueError('go away you bad dude')

        setattr(cls, 'transform', transform)
        setattr(cls, 'resolve', resolve)
        setattr(cls, 'from_tokens', from_tokens)
        setattr(cls, '__init__', __init__)


class TextMatrix(metaclass=MatrixCommand):
    COMMAND = 'Tm'


class CTM(metaclass=MatrixCommand):
    COMMAND = 'cm'


def format_number(n):
    # Really small numbers should just be rounded to 0
    if -ZERO_TOLERANCE <= n <= ZERO_TOLERANCE:
        return '0'
    # Cut off unnecessary decimals
    if n % 1 == 0:
        return str(int(n))
    # Otherwise return 10 decimal places, but remove trailing zeros. I wish I
    # could use 'g' for this, but that switches to scientific notation at
    # certain thresholds.
    string = '{:.10f}'.format(n)
    i = len(string) - 1
    while string[i] == '0':
        i -= 1
    return string[:i + 1]


def quadratic_to_cubic_bezier(
    start_x, start_y,
    control_x, control_y,
    end_x, end_y
):
    """Make a cubic bezier curve from the parameters of a quadratic bezier.

    This is necessary because PDF doesn't have quadratic beziers.
    """
    cp1x = start_x + 2 / 3 * (control_x - start_x)
    cp1y = start_y + 2 / 3 * (control_y - start_y)
    cp2x = end_x + 2 / 3 * (control_x - end_x)
    cp2y = end_y + 2 / 3 * (control_y - end_y)
    return Bezier(cp1x, cp1y, cp2x, cp2y, end_x, end_y)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"FPDF for python"

__license__ = "LGPL 3.0"
__version__ = "1.7.2"

from .fpdf import FPDF, FPDF_FONT_DIR, FPDF_VERSION, SYSTEM_TTFONTS, set_global, FPDF_CACHE_MODE, FPDF_CACHE_DIR
try:
    from .html import HTMLMixin
except ImportError:
    import warnings
    warnings.warn("web2py gluon package not installed, required for html2pdf")

from .template import Template



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
import pprint
import binascii
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.serialization import pkcs12
from asn1crypto import x509, pem


def cert2asn(cert_bytes):
    if pem.detect(cert_bytes):
        _, _, cert_bytes = pem.unarmor(cert_bytes)
    return x509.Certificate.load(cert_bytes)

def main():
    if len(sys.argv) == 2:
        pemname = sys.argv[1]
    else:
        pemname = 'ca/demo2_ca.crt.pem'
    data = open(pemname, 'rb').read()
    cert = cert2asn(data)
    pprint.pprint(cert.native)
main()



#!/usr/bin/env vpython3
"""Extract pdf structure in XML format"""
import logging
import os.path
import re
import sys
from argparse import ArgumentParser

from pdfminer.pdfdocument import PDFDocument, PDFNoOutlines
from pdfminer.pdfpage import PDFPage
from pdfminer.pdfparser import PDFParser
from pdfminer.pdftypes import PDFObjectNotFound, PDFValueError
from pdfminer.pdftypes import PDFStream, PDFObjRef, resolve1, stream_value
from pdfminer.psparser import PSKeyword, PSLiteral, LIT
from pdfminer.utils import isnumber

logging.basicConfig()

ESC_PAT = re.compile(r'[\000-\037&<>()"\042\047\134\177-\377]')


def e(s):
    if isinstance(s, bytes):
        s = str(s, "latin-1")
    return ESC_PAT.sub(lambda m: "&#%d;" % ord(m.group(0)), s)


def dumpxml(out, obj, codec=None):
    if obj is None:
        out.write("<null />")
        return

    if isinstance(obj, dict):
        out.write('<dict size="%d">\n' % len(obj))
        for (k, v) in obj.items():
            out.write("<key>%s</key>\n" % k)
            out.write("<value>")
            dumpxml(out, v)
            out.write("</value>\n")
        out.write("</dict>")
        return

    if isinstance(obj, list):
        out.write('<list size="%d">\n' % len(obj))
        for v in obj:
            dumpxml(out, v)
            out.write("\n")
        out.write("</list>")
        return

    if isinstance(obj, ((str,), bytes)):
        out.write('<string size="%d">%s</string>' % (len(obj), e(obj)))
        return

    if isinstance(obj, PDFStream):
        if codec == "raw":
            out.write(obj.get_rawdata())
        elif codec == "binary":
            out.write(obj.get_data())
        else:
            out.write("<stream>\n<props>\n")
            dumpxml(out, obj.attrs)
            out.write("\n</props>\n")
            if codec == "text":
                data = obj.get_data()
                out.write('<data size="%d">%s</data>\n' % (len(data), e(data)))
            out.write("</stream>")
        return

    if isinstance(obj, PDFObjRef):
        out.write('<ref id="%d" />' % obj.objid)
        return

    if isinstance(obj, PSKeyword):
        out.write("<keyword>%s</keyword>" % obj.name)
        return

    if isinstance(obj, PSLiteral):
        out.write("<literal>%s</literal>" % obj.name)
        return

    if isnumber(obj):
        out.write("<number>%s</number>" % obj)
        return

    raise TypeError(obj)


def dumptrailers(out, doc):
    for xref in doc.xrefs:
        out.write("<trailer>\n")
        dumpxml(out, xref.trailer)
        out.write("\n</trailer>\n\n")
    return


def dumpallobjs(out, doc, codec=None):
    visited = set()
    out.write("<pdf>")
    for xref in doc.xrefs:
        for objid in xref.get_objids():
            if objid in visited:
                continue
            visited.add(objid)
            try:
                obj = doc.getobj(objid)
                if obj is None:
                    continue
                out.write('<object id="%d">\n' % objid)
                dumpxml(out, obj, codec=codec)
                out.write("\n</object>\n\n")
            except PDFObjectNotFound as e:
                print("not found: %r" % e)
    dumptrailers(out, doc)
    out.write("</pdf>")
    return


def dumpoutline(
    outfp,
    fname,
    objids,
    pagenos,
    password="",
    dumpall=False,
    codec=None,
    extractdir=None,
):
    fp = open(fname, "rb")
    parser = PDFParser(fp)
    doc = PDFDocument(parser, password)
    pages = {
        page.pageid: pageno
        for (pageno, page) in enumerate(PDFPage.create_pages(doc), 1)
    }

    def resolve_dest(dest):
        if isinstance(dest, str):
            dest = resolve1(doc.get_dest(dest))
        elif isinstance(dest, PSLiteral):
            dest = resolve1(doc.get_dest(dest.name))
        if isinstance(dest, dict):
            dest = dest["D"]
        if isinstance(dest, PDFObjRef):
            dest = dest.resolve()
        return dest

    try:
        outlines = doc.get_outlines()
        outfp.write("<outlines>\n")
        for (level, title, dest, a, se) in outlines:
            pageno = None
            if dest:
                dest = resolve_dest(dest)
                pageno = pages[dest[0].objid]
            elif a:
                action = a
                if isinstance(action, dict):
                    subtype = action.get("S")
                    if subtype and repr(subtype) == "/'GoTo'" and action.get("D"):
                        dest = resolve_dest(action["D"])
                        pageno = pages[dest[0].objid]
            s = e(title).encode("utf-8", "xmlcharrefreplace")
            outfp.write('<outline level="{!r}" title="{}">\n'.format(level, s))
            if dest is not None:
                outfp.write("<dest>")
                dumpxml(outfp, dest)
                outfp.write("</dest>\n")
            if pageno is not None:
                outfp.write("<pageno>%r</pageno>\n" % pageno)
            outfp.write("</outline>\n")
        outfp.write("</outlines>\n")
    except PDFNoOutlines:
        pass
    parser.close()
    fp.close()
    return


LITERAL_FILESPEC = LIT("Filespec")
LITERAL_EMBEDDEDFILE = LIT("EmbeddedFile")


def extractembedded(
    outfp,
    fname,
    objids,
    pagenos,
    password="",
    dumpall=False,
    codec=None,
    extractdir=None,
):
    def extract1(objid, obj):
        filename = os.path.basename(obj.get("UF") or obj.get("F").decode())
        fileref = obj["EF"].get("UF") or obj["EF"].get("F")
        fileobj = doc.getobj(fileref.objid)
        if not isinstance(fileobj, PDFStream):
            error_msg = (
                "unable to process PDF: reference for %r is not a "
                "PDFStream" % filename
            )
            raise PDFValueError(error_msg)
        if fileobj.get("Type") is not LITERAL_EMBEDDEDFILE:
            raise PDFValueError(
                "unable to process PDF: reference for %r "
                "is not an EmbeddedFile" % (filename)
            )
        path = os.path.join(extractdir, "%.6d-%s" % (objid, filename))
        if os.path.exists(path):
            raise IOError("file exists: %r" % path)
        print("extracting: %r" % path)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        out = open(path, "wb")
        out.write(fileobj.get_data())
        out.close()
        return

    with open(fname, "rb") as fp:
        parser = PDFParser(fp)
        doc = PDFDocument(parser, password)
        extracted_objids = set()
        for xref in doc.xrefs:
            for objid in xref.get_objids():
                obj = doc.getobj(objid)
                if (
                    objid not in extracted_objids
                    and isinstance(obj, dict)
                    and obj.get("Type") is LITERAL_FILESPEC
                ):
                    extracted_objids.add(objid)
                    extract1(objid, obj)
    return


def dumppdf(
    outfp,
    fname,
    objids,
    pagenos,
    password="",
    dumpall=False,
    codec=None,
    extractdir=None,
):
    fp = open(fname, "rb")
    parser = PDFParser(fp)
    doc = PDFDocument(parser, password)
    if objids:
        for objid in objids:
            obj = doc.getobj(objid)
            dumpxml(outfp, obj, codec=codec)
    if pagenos:
        for (pageno, page) in enumerate(PDFPage.create_pages(doc)):
            if pageno in pagenos:
                if codec:
                    for obj in page.contents:
                        obj = stream_value(obj)
                        dumpxml(outfp, obj, codec=codec)
                else:
                    dumpxml(outfp, page.attrs)
    if dumpall:
        dumpallobjs(outfp, doc, codec=codec)
    if (not objids) and (not pagenos) and (not dumpall):
        dumptrailers(outfp, doc)
    fp.close()
    if codec not in ("raw", "binary"):
        outfp.write("\n")
    return


def create_parser():
    parser = ArgumentParser(description=__doc__, add_help=True)
    parser.add_argument(
        "files",
        type=str,
        default=None,
        nargs="+",
        help="One or more paths to PDF files.",
    )

    parser.add_argument(
        "--debug",
        "-d",
        default=False,
        action="store_true",
        help="Use debug logging level.",
    )
    procedure_parser = parser.add_mutually_exclusive_group()
    procedure_parser.add_argument(
        "--extract-toc",
        "-T",
        default=False,
        action="store_true",
        help="Extract structure of outline",
    )
    procedure_parser.add_argument(
        "--extract-embedded", "-E", type=str, help="Extract embedded files"
    )

    parse_params = parser.add_argument_group(
        "Parser", description="Used during PDF parsing"
    )
    parse_params.add_argument(
        "--page-numbers",
        type=int,
        default=None,
        nargs="+",
        help="A space-seperated list of page numbers to parse.",
    )
    parse_params.add_argument(
        "--pagenos",
        "-p",
        type=str,
        help="A comma-separated list of page numbers to parse. Included for "
        "legacy applications, use --page-numbers for more idiomatic "
        "argument entry.",
    )
    parse_params.add_argument(
        "--objects",
        "-i",
        type=str,
        help="Comma separated list of object numbers to extract",
    )
    parse_params.add_argument(
        "--all",
        "-a",
        default=False,
        action="store_true",
        help="If the structure of all objects should be extracted",
    )
    parse_params.add_argument(
        "--password",
        "-P",
        type=str,
        default="",
        help="The password to use for decrypting PDF file.",
    )

    output_params = parser.add_argument_group(
        "Output", description="Used during output generation."
    )
    output_params.add_argument(
        "--outfile",
        "-o",
        type=str,
        default="-",
        help='Path to file where output is written. Or "-" (default) to '
        "write to stdout.",
    )
    codec_parser = output_params.add_mutually_exclusive_group()
    codec_parser.add_argument(
        "--raw-stream",
        "-r",
        default=False,
        action="store_true",
        help="Write stream objects without encoding",
    )
    codec_parser.add_argument(
        "--binary-stream",
        "-b",
        default=False,
        action="store_true",
        help="Write stream objects with binary encoding",
    )
    codec_parser.add_argument(
        "--text-stream",
        "-t",
        default=False,
        action="store_true",
        help="Write stream objects as plain text",
    )

    return parser


def main(argv=None):
    parser = create_parser()
    args = parser.parse_args(args=argv)

    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    if args.outfile == "-":
        outfp = sys.stdout
    else:
        outfp = open(args.outfile, "w")

    if args.objects:
        objids = [int(x) for x in args.objects.split(",")]
    else:
        objids = []

    if args.page_numbers:
        pagenos = {x - 1 for x in args.page_numbers}
    elif args.pagenos:
        pagenos = {int(x) - 1 for x in args.pagenos.split(",")}
    else:
        pagenos = set()

    password = args.password

    if args.raw_stream:
        codec = "raw"
    elif args.binary_stream:
        codec = "binary"
    elif args.text_stream:
        codec = "text"
    else:
        codec = None

    if args.extract_toc:
        extractdir = None
        proc = dumpoutline
    elif args.extract_embedded:
        extractdir = args.extract_embedded
        proc = extractembedded
    else:
        extractdir = None
        proc = dumppdf

    for fname in args.files:
        proc(
            outfp,
            fname,
            objids,
            pagenos,
            password=password,
            dumpall=args.all,
            codec=codec,
            extractdir=extractdir,
        )
    outfp.close()
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
import datetime
import base64
import json
import hashlib
from asn1crypto import cms, core, util
from endesive import pdf


class Signer:
    def __init__(self, cert, sig, tosign):
        self.cert = cert
        self.sig = sig
        self.tosign = tosign
        self.mech = None

    def certificate(self):
        return 1, self.cert

    def sign(self, keyid, data, mech):
        if self.tosign:
            assert self.tosign == data
        self.tosign = data
        self.mech = mech
        if self.sig is None:
            sig = None
            if mech == "sha256":
                sig = b"\0" * 256
            return sig
        return self.sig


def main():
    def attrs(signed_value):
        result = [
            cms.CMSAttribute(
                {"type": cms.CMSAttributeType("content_type"), "values": ("data",)}
            ),
            cms.CMSAttribute(
                {
                    "type": cms.CMSAttributeType("message_digest"),
                    "values": (signed_value,),
                }
            ),
            cms.CMSAttribute(
                {
                    "type": cms.CMSAttributeType("signing_time"),
                    "values": (cms.Time({"utc_time": core.UTCTime(signed_time)}),),
                }
            ),
        ]
        return result

    dct = {
        "sigflags": 3,
        "sigpage": 0,
        "sigbutton": True,
        "contact": "mak@trisoft.com.pl",
        "location": "Szczecin",
        "reason": "Dokument podpisany cyfrowo",
        "signature": "Dokument podpisany cyfrowo",
        "signaturebox": (0, 0, 100, 100),
        "aligned": 16384,
        "attrs": attrs,
        "newid": "1",
    }

    pdfname = 'pdf.pdf'
    if len (sys.argv) > 1:
        pdfname = sys.argv[1]
    try:
        config = open(pdfname + ".json", "rt").read()
        config = json.loads(config)
    except FileNotFoundError:
        when = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        config = {
            'when': when,
            'certificate': open("cert-hsm-user1.pem", "rt").read(),
            'signed_bytes': None,
            'tosign': None,
            'id': hashlib.md5(when.encode()).hexdigest()
        }
    dct['id'] = config['id'].encode()
    when = datetime.datetime.strptime(config['when'], "%Y-%m-%d %H:%M:%S")
    dct["signingdate"] = when.strftime("%Y%m%d%H%M%S+00'00'").encode()
    signed_time = datetime.datetime(
        when.year, when.month, when.day, when.hour, when.minute, when.second, 0, util.timezone.utc
    )
    cert = config['certificate'].encode('ascii')
    signed_bytes = config['signed_bytes']
    if signed_bytes is not None:
        signed_bytes = base64.decodebytes(signed_bytes.encode('ascii'))
    tosign = config['tosign']
    if tosign is not None:
        tosign = base64.decodebytes(tosign.encode('ascii'))

    clshsm = Signer(cert, signed_bytes, tosign)

    datau = open(pdfname, "rb").read()
    cls = pdf.cms.SignedData()
    datas = cls.sign(datau, dct, None, None, [], "sha256", clshsm, mode="sign")

    if signed_bytes is None:
        config['tosign'] = b"".join(base64.encodebytes(clshsm.tosign).split()).decode('ascii')
        json.dump(config, open(pdfname + ".json", "wt"), indent=4)
    else:
        fname = pdfname.replace(".pdf", "-signed-cms-hash.pdf")
        with open(fname, "wb") as fp:
            fp.write(datau)
            fp.write(datas)


main()



# -*- coding: utf-8 -*-
"""
    Annotation
    ~~~~~~~~~~~~
    Base annotation class.

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
from ..pdfrw import PdfDict, PdfArray, PdfName, IndirectPdfDict

from ..config.constants import GRAPHICS_STATE_NAME
from ..config.metadata import serialize_value
from ..util.geometry import transform_rect
from ..util.geometry import translate


ALL_VERSIONS = ("1.3", "1.4", "1.5", "1.6", "1.7")


class Annotation(object):
    """Base class for all PDF annotation objects.

    Concrete annotations should define the following:
        * subtype (e.g. "Square")
        * make_rect() - bounding box of annotation
        * add_additional_pdf_object_data [optional] - additional entries to go
          in the PDF object
        * add_additional_resources [optional] - additional entries to go in the
          Resources sub-dict of the annotation

    There is a lot of nuance and viewer-specific (mostly Acrobat and Bluebeam)
    details to consider when creating PDF annotations. One big thing that's not
    immediately clear from the PDF spec is that wherever possible, we fill in
    the annotations' type-specific details (e.g. BE and IC for squares), but
    also create and include an Appearance Stream. The latter gives us control
    over exactly how the annotation appears across different viewers, while the
    former allows Acrobat or BB to regenerate the appearance stream during
    editing.
    """

    versions = ALL_VERSIONS

    def __init__(self, location, appearance, metadata=None):
        """
        :param Location location:
        :param Appearance appearance:
        :param Metadata metadata:
        """
        self._location = location
        self._appearance = appearance
        self._metadata = metadata

    def as_pdf_object(self, transform, page):
        """Return the PdfDict object representing the annotation, that will be
        inserted as is into the PDF document.

        :param list transform: Transformation matrix to transform the coords
            of the annotation from client-specified space to PDF user space.
        :param PdfDict page: The pdfrw page object from the PDF document
        :returns PdfDict: the annotation object to be inserted into the PDF
        """
        bounding_box = transform_rect(self.make_rect(), transform)
        appearance_stream = self._make_appearance_stream_dict(bounding_box, transform)

        obj = PdfDict(
            Type=PdfName("Annot"),
            Subtype=PdfName(self.subtype),
            Rect=bounding_box,
            AP=appearance_stream,
            P=page,
        )

        self._add_metadata(obj, self._metadata)
        self.add_additional_pdf_object_data(obj)
        obj.indirect = True

        return obj

    @property
    def page(self):
        return self._location.page

    def validate(self, pdf_version):
        """Validate a new annotation against a given PDF version."""
        pass

    def _add_metadata(self, obj, metadata):
        if metadata is None:
            return
        for name, value in metadata.iter():
            obj[PdfName(name)] = serialize_value(value)

    def _make_ap_resources(self):
        """Make the Resources entry for the appearance stream dictionary.

        Implement add_additional_resources to add additional entries -
        fonts, XObjects, graphics state - to the Resources dictionary.
        """
        resources = IndirectPdfDict(ProcSet=PdfArray([PdfName("PDF")]))
        self._add_graphics_state_resources(resources, self._appearance)
        self._add_xobject_resources(resources, self._appearance)
        self._add_font_resources(resources, self._appearance)
        self.add_additional_resources(resources)
        return resources

    @staticmethod
    def _add_font_resources(resources, A):
        if A.fonts:
            resources.Font = IndirectPdfDict()
            for font_name, font in A.fonts.items():
                resources.Font[PdfName(font_name)] = font

    @staticmethod
    def _add_xobject_resources(resources, A):
        """Adds in provided, explicit XObjects into the appearance stream's
        Resources dict. This is used when the user is explicitly specifying the
        appearance stream and they want to include, say, an image.
        """
        if A.xobjects:
            resources.XObject = IndirectPdfDict()
            for xobject_name, xobject in A.xobjects.items():
                resources.XObject[PdfName(xobject_name)] = xobject

    @staticmethod
    def _add_graphics_state_resources(resources, A):
        """Add in the resources dict for turning on transparency in the
        graphics state. For example, if both stroke and fill were transparent,
        this would add:
            << /ExtGState /PdfAnnotatorGS <<
                /CA 0.5 /ca 0.75 /Type /ExtGState
            >> >>
        to the Resources dict.

        Graphics states can also be specified externally, for use in explicit
        content streams. This is done by using the `graphics_states` property
        on the appearance object.
        """
        states = []
        internal_state = Annotation._get_internal_graphics_state(resources, A)
        if internal_state is not None:
            states.append((GRAPHICS_STATE_NAME, internal_state))

        if A.graphics_states:
            for name, state in A.graphics_states.items():
                states.append((name, state.as_pdf_dict()))

        if states:
            resources.ExtGState = PdfDict()
            for name, state in states:
                resources.ExtGState[PdfName(name)] = state

    @staticmethod
    def _get_internal_graphics_state(resources, A):
        internal_state = A.get_graphics_state()
        if internal_state.has_content():
            return internal_state.as_pdf_dict()

        return None

    def _make_appearance_stream_dict(self, bounding_box, transform):

        # Either use user-specified content stream or generate content stream
        # based on annotation type.
        stream = self._appearance.appearance_stream
        if stream is None:
            stream = self.make_appearance_stream()

        resources = self._make_ap_resources()

        # Transform the appearance stream into PDF space and turn it into a str
        appearance_stream = stream.transform(transform).resolve()

        normal_appearance = IndirectPdfDict(
            stream=appearance_stream,
            BBox=bounding_box,
            Resources=resources,
            Matrix=translate(-bounding_box[0], -bounding_box[1]),
            Type=PdfName("XObject"),
            Subtype=PdfName("Form"),
            FormType=1,
        )
        return PdfDict(N=normal_appearance)

    def add_additional_pdf_object_data(self, obj):
        """Add additional keys to the PDF object. Default is a no-op.

        :param PdfDict obj: the PDF object to be inserted into the PDF
        """
        pass

    def add_additional_resources(self, resources):
        """Add additional keys to the Resources PDF dictionary. Default is a
        no-op.

        :param PdfDict resources: Resources PDF dictionary
        """
        pass

    def make_rect(self):
        """Return a bounding box that encompasses the entire annotation."""
        raise NotImplementedError()


def make_border_dict(appearance):
    A = appearance
    return _make_border_dict(A.stroke_width, A.border_style, A.dash_array)


def _make_border_dict(width, style, dash_array=None):
    border = PdfDict(Type=PdfName("Border"), W=width, S=PdfName(style))
    if dash_array:
        if style != "D":
            raise ValueError("Dash array only applies to dashed borders!")
        border.D = dash_array
    return border


class Stamp(object):
    subtype = "Stamp"



# -*- coding: utf-8 -*-
# https://dss.nowina.lu/validation
# https://signatures-conformance-checker.etsi.org/pub/index.php
#
import base64
from endesive import xades
import time
import datetime
import hashlib
import io
import uuid

from cryptography.x509.oid import NameOID
from lxml import etree, builder
import requests
from asn1crypto import cms, algos, core, keys, pem, tsp, x509, util

DS = builder.ElementMaker(
    namespace="http://www.w3.org/2000/09/xmldsig#",
    nsmap={"ds": "http://www.w3.org/2000/09/xmldsig#"},
)
CanonicalizationMethod = DS.CanonicalizationMethod
DigestMethod = DS.DigestMethod
DigestValue = DS.DigestValue
KeyInfo = DS.KeyInfo
Object = DS.Object
Reference = DS.Reference
Signature = DS.Signature
SignatureMethod = DS.SignatureMethod
SignatureValue = DS.SignatureValue
SignedInfo = DS.SignedInfo
Transform = DS.Transform
Transforms = DS.Transforms
X509Certificate = DS.X509Certificate
X509Data = DS.X509Data
X509IssuerName = DS.X509IssuerName
X509SerialNumber = DS.X509SerialNumber
XPath = DS.XPath

XADES = builder.ElementMaker(
    namespace="http://uri.etsi.org/01903/v1.3.2#",
    nsmap={
        "xades": "http://uri.etsi.org/01903/v1.3.2#",
        "ds": "http://www.w3.org/2000/09/xmldsig#",
    },
)
Cert = XADES.Cert
CertDigest = XADES.CertDigest
DataObjectFormat = XADES.DataObjectFormat
Description = XADES.Description
DocumentationReference = XADES.DocumentationReference
DocumentationReferences = XADES.DocumentationReferences
Identifier = XADES.Identifier
IssuerSerial = XADES.IssuerSerial
MimeType = XADES.MimeType
ObjectIdentifier = XADES.ObjectIdentifier
QualifyingProperties = XADES.QualifyingProperties
SignedDataObjectProperties = XADES.SignedDataObjectProperties
SignedProperties = XADES.SignedProperties
SignedSignatureProperties = XADES.SignedSignatureProperties
SigningCertificate = XADES.SigningCertificate
SigningTime = XADES.SigningTime
UnsignedProperties = XADES.UnsignedProperties
UnsignedSignatureProperties = XADES.UnsignedSignatureProperties
SignatureTimeStamp = XADES.SignatureTimeStamp
EncapsulatedTimeStamp = XADES.EncapsulatedTimeStamp


OID_NAMES = {
    NameOID.COMMON_NAME: "CN",
    NameOID.COUNTRY_NAME: "C",
    NameOID.DOMAIN_COMPONENT: "DC",
    NameOID.EMAIL_ADDRESS: "E",
    NameOID.GIVEN_NAME: "G",
    NameOID.LOCALITY_NAME: "L",
    NameOID.ORGANIZATION_NAME: "O",
    NameOID.ORGANIZATIONAL_UNIT_NAME: "OU",
    NameOID.SURNAME: "SN",
}


class BES:
    debug = False

    def __init__(self):
        self.guid = str(uuid.uuid1())
        self.time = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")

    def sha256(self, data):
        h = hashlib.sha256(data).digest()
        return base64.b64encode(h).decode()

    def base64(self, data):
        b64 = b"".join(base64.encodebytes(data).split())
        data = []
        for i in range(0, len(b64), 64):
            data.append(b64[i : i + 64])
        data = b"\n".join(data).decode()
        return data

    def get_rdns_name(self, rdns):
        name = ""
        for rdn in rdns:
            for attr in rdn._attributes:
                if len(name) > 0:
                    name = name + ","
                if attr.oid in OID_NAMES:
                    name = name + OID_NAMES[attr.oid]
                else:
                    name = name + attr.oid.dotted_string
                    s = "".join(["%02x" % int(b) for b in attr.value.encode()])
                    s = "#0C%02X%s" % (len(attr.value), s)
                    name = name + "=" + s
                    continue
                name = name + "=" + attr.value
        return name

    def _c14n(self, nodes, algorithm, inclusive_ns_prefixes=None):
        exclusive, with_comments = False, False

        if algorithm.startswith("http://www.w3.org/2001/10/xml-exc-c14n#"):
            exclusive = True
        if algorithm.endswith("#WithComments"):
            with_comments = True

        data = etree.tostring(
            nodes, encoding="UTF-8", xml_declaration=True, standalone=False
        )
        data = io.BytesIO(data)
        tree = etree.parse(data)
        data = io.BytesIO()
        tree.write_c14n(
            data,
            exclusive=exclusive,
            with_comments=with_comments,
            compression=0,
            inclusive_ns_prefixes=inclusive_ns_prefixes,
        )
        c14n = data.getvalue()

        return c14n

    def unsignedpropertied(self, signed_value, tspurl, tspcred, hashalgo="sha256"):
        if tspurl is None:
            unsignedproperties = UnsignedProperties(
                Id="UnsignedProperties_" + self.guid + self.mapa["_5d"]
            )
        else:
            tspreq = tsp.TimeStampReq(
                {
                    "version": 1,
                    "message_imprint": tsp.MessageImprint(
                        {
                            "hash_algorithm": algos.DigestAlgorithm(
                                {"algorithm": hashalgo}
                            ),
                            "hashed_message": signed_value.encode(),
                        }
                    ),
                    #'req_policy', ObjectIdentifier, {'optional': True}),
                    "nonce": int(time.time() * 1000),
                    "cert_req": True,
                    #'extensions': tsp.Extensions()
                }
            )
            tspreq = tspreq.dump()

            tspheaders = {"Content-Type": "application/timestamp-query"}
            if tspcred is not None:
                username = tspcred.get("username", None)
                password = tspcred.get("password", None)
                if username and password:
                    auth_header_value = base64.b64encode(
                        bytes(username + ":" + password, "utf-8")
                    ).decode("ascii")
                    tspheaders["Authorization"] = f"Basic {auth_header_value}"
            tspresp = requests.post(tspurl, data=tspreq, headers=tspheaders)
            if (
                tspresp.headers.get("Content-Type", None)
                == "application/timestamp-reply"
            ):
                tspresp = tsp.TimeStampResp.load(tspresp.content)

                if tspresp["status"]["status"].native == "granted":
                    attr = self.base64(tspresp["time_stamp_token"].dump())
                else:
                    raise ValueError("TimeStampResponse status is not granted")
            else:
                raise ValueError("TimeStampResponse has invalid content type")

            unsignedproperties = UnsignedProperties(
                UnsignedSignatureProperties(
                    SignatureTimeStamp(
                        CanonicalizationMethod(
                            Algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315"
                        ),
                        EncapsulatedTimeStamp(
                            attr, Encoding="http://uri.etsi.org/01903/v1.2.2#DER"
                        ),
                        Id="SignatureTimeStamp_" + self.guid,
                    )
                ),
                Id="UnsignedProperties_" + self.guid + self.mapa["_5d"],
            )
        return unsignedproperties

    mapa = {
        "_02": "_5d",
        "_2f": "_70",
        "_43": "_1c",
        "_20": "_7f",
        "_46": "_19",
        "_49": "_16",
        "_5a": "_05",
        "_2c": "_73",
        "_4b": "_14",
        "_11": "_4e",
        "_5d": "_02",
    }

    def enveloped(self, data, cert, certcontent, signproc, tspurl, tspcred):
        tree = etree.parse(io.BytesIO(data))
        signedobj = tree.getroot()
        canonicalizedxml = self._c14n(signedobj, "")
        digestvalue1 = self.sha256(canonicalizedxml)

        nsmap = signedobj.nsmap.copy()
        nsmap.update(
            {
                "xades": "http://uri.etsi.org/01903/v1.3.2#",
                "ds": "http://www.w3.org/2000/09/xmldsig#",
            }
        )
        siXADES = builder.ElementMaker(
            namespace="http://uri.etsi.org/01903/v1.3.2#", nsmap=nsmap
        )
        SignedProperties = siXADES.SignedProperties

        nsmap = signedobj.nsmap.copy()
        nsmap.update({"ds": "http://www.w3.org/2000/09/xmldsig#"})
        siDS = builder.ElementMaker(
            namespace="http://www.w3.org/2000/09/xmldsig#", nsmap=nsmap
        )
        SignedInfo = siDS.SignedInfo

        certdigest = self.sha256(certcontent)
        certcontent = self.base64(certcontent)
        certserialnumber = "%d" % cert.serial_number
        certissuer = self.get_rdns_name(cert.issuer.rdns)
        if self.debug:
            self.guid = "279d6285-779c-4449-9c92-6bf3f7edacc2"
            self.time = "2020-11-24T00:32:43Z"
            certissuer = "2.5.4.97=#0C10564154504C2D35313730333539343538,CN=Certum QCA 2017,O=Asseco Data Systems S.A.,C=PL"

        signedproperties = SignedProperties(
            SignedSignatureProperties(
                SigningTime(self.time),
                SigningCertificate(
                    Cert(
                        CertDigest(
                            DigestMethod(
                                Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"
                            ),
                            DigestValue(certdigest),
                        ),
                        IssuerSerial(
                            X509IssuerName(certissuer),
                            X509SerialNumber(certserialnumber),
                        ),
                    )
                ),
                Id="SignedSignatureProperties_" + self.guid + self.mapa["_02"],
            ),
            SignedDataObjectProperties(
                DataObjectFormat(
                    Description(
                        """\
MIME-Version: 1.0
Content-Type: text/xml
Content-Transfer-Encoding: binary
Content-Disposition: filename="document.xml"\
"""
                    ),
                    ObjectIdentifier(
                        Identifier(
                            "http://www.certum.pl/OIDAsURI/signedFile/1.2.616.1.113527.3.1.1.3.1",
                            Qualifier="OIDAsURI",
                        ),
                        Description("Opis formatu dokumentu oraz jego pełna nazwa"),
                        DocumentationReferences(
                            DocumentationReference(
                                "http://www.certum.pl/OIDAsURI/signedFile.pdf"
                            )
                        ),
                    ),
                    MimeType("text/xml"),
                    ObjectReference="#Reference1_" + self.guid + self.mapa["_2f"],
                ),
                Id="SignedDataObjectProperties_" + self.guid + self.mapa["_43"],
            ),
            Id="SignedProperties_" + self.guid + self.mapa["_46"],
        )

        canonicalizedxml = self._c14n(signedproperties, "")
        digestvalue2 = self.sha256(canonicalizedxml)
        if self.debug:
            print("*" * 20, "enveloped signedproperties")
            print(canonicalizedxml)
            print("digest:", digestvalue2)

        unsignedproperties = self.unsignedpropertied(
            digestvalue2, tspurl, tspcred, "sha256"
        )

        signedinfo = SignedInfo(
            CanonicalizationMethod(
                Algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315"
            ),
            SignatureMethod(
                Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"
            ),
            Reference(
                Transforms(
                    Transform(
                        XPath("not(ancestor-or-self::ds:Signature)"),
                        Algorithm="http://www.w3.org/TR/1999/REC-xpath-19991116",
                    )
                ),
                DigestMethod(Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"),
                DigestValue(digestvalue1),
                Id="Reference1_" + self.guid + self.mapa["_2f"],
                URI="",
            ),
            Reference(
                DigestMethod(Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"),
                DigestValue(digestvalue2),
                Id="SignedProperties-Reference_" + self.guid + self.mapa["_20"],
                Type="http://uri.etsi.org/01903#SignedProperties",
                URI="#SignedProperties_" + self.guid + self.mapa["_46"],
            ),
            Id="SignedInfo_" + self.guid + self.mapa["_49"],
        )

        canonicalizedxml = self._c14n(signedinfo, "")
        if self.debug:
            print("*" * 20, "enveloped signedinfo")
            print(canonicalizedxml)
        signature = signproc(canonicalizedxml, "sha256")
        actualdigestencoded = base64.b64encode(signature).decode()
        digestvalue3 = []
        for i in range(0, len(actualdigestencoded), 64):
            digestvalue3.append(actualdigestencoded[i : i + 64])
        digestvalue3 = "\n".join(digestvalue3)

        DOC = Signature(
            signedinfo,
            SignatureValue(
                digestvalue3, Id="SignatureValue_" + self.guid + self.mapa["_5a"]
            ),
            KeyInfo(
                X509Data(X509Certificate(certcontent)),
                Id="KeyInfo_" + self.guid + self.mapa["_2c"],
            ),
            Object(
                QualifyingProperties(
                    signedproperties,
                    unsignedproperties,
                    Id="QualifyingProperties_" + self.guid + self.mapa["_4b"],
                    Target="#Signature_" + self.guid + self.mapa["_11"],
                )
            ),
            Id="Signature_" + self.guid + self.mapa["_11"],
        )

        signedobj.append(DOC)
        return tree

    def enveloping(
        self,
        fname,
        data,
        smime,
        cert,
        certcontent,
        signproc,
        base64encode=True,
        withcomments=False,
        detached=False,
        tspurl=None,
        tspcred=None,
    ):
        swithcomments = ""
        if withcomments:
            swithcomments = "#WithComments"
        if detached:
            tree = etree.parse(io.BytesIO(data))
            signedobj = tree.getroot()
            canonicalizedxml = self._c14n(signedobj, "")
            digestvalue1 = self.sha256(canonicalizedxml)
            URI = fname
            signedobj = None
        else:
            if base64encode:
                data = base64.b64encode(data).decode()
                signedobj = Object(
                    data,
                    Encoding="http://www.w3.org/2000/09/xmldsig#base64",
                    MimeType=smime,
                    Id="Object1_" + self.guid,
                )
                URI = "#Object1_" + self.guid
            elif 0:
                signedobj = Object(data, MimeType="text/xml", Id="Object1_" + self.guid)
                URI = "#Object1_" + self.guid
            else:
                signedobj = Object(MimeType="text/xml", Id="Object1_" + self.guid)
                tree = etree.parse(io.BytesIO(data))
                signedobj.append(tree.getroot())
                URI = "#Object1_" + self.guid
            canonicalizedxml = self._c14n(signedobj, "")
            digestvalue1 = self.sha256(canonicalizedxml)

        certdigest = self.sha256(certcontent)
        certcontent = self.base64(certcontent)
        certserialnumber = "%d" % cert.serial_number
        certissuer = self.get_rdns_name(cert.issuer.rdns)

        signedprop = SignedProperties(
            SignedSignatureProperties(
                SigningTime(self.time),
                SigningCertificate(
                    Cert(
                        CertDigest(
                            DigestMethod(
                                Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"
                            ),
                            DigestValue(certdigest),
                        ),
                        IssuerSerial(
                            X509IssuerName(certissuer),
                            X509SerialNumber(certserialnumber),
                        ),
                    )
                ),
                Id="SignedSignatureProperties_" + self.guid + self.mapa["_02"],
            ),
            SignedDataObjectProperties(
                DataObjectFormat(
                    Description(
                        """\
MIME-Version: 1.0
Content-Type: %s
Content-Transfer-Encoding: binary
Content-Disposition: filename="%s"\
"""
                        % (smime, fname)
                    ),
                    ObjectIdentifier(
                        Identifier(
                            "http://www.certum.pl/OIDAsURI/signedFile/1.2.616.1.113527.3.1.1.3.1",
                            Qualifier="OIDAsURI",
                        ),
                        Description("Opis formatu dokumentu oraz jego pełna nazwa"),
                        DocumentationReferences(
                            DocumentationReference(
                                "http://www.certum.pl/OIDAsURI/signedFile.pdf"
                            )
                        ),
                    ),
                    MimeType(smime),
                    ObjectReference="#Reference1_" + self.guid + self.mapa["_2f"],
                ),
                Id="SignedDataObjectProperties_" + self.guid + self.mapa["_43"],
            ),
            Id="SignedProperties_" + self.guid + self.mapa["_46"],
        )

        canonicalizedxml = self._c14n(signedprop, "")
        digestvalue2 = self.sha256(canonicalizedxml)
        if self.debug:
            print("*" * 20, "build signedprop")
            print(canonicalizedxml)
            print("digest:", digestvalue2)

        signedinfo = SignedInfo(
            CanonicalizationMethod(
                Algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315"
            ),
            SignatureMethod(
                Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"
            ),
            Reference(
                Transforms(
                    Transform(
                        Algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315"
                        + swithcomments
                    )
                ),
                DigestMethod(Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"),
                DigestValue(digestvalue1),
                URI=URI,
                Id="Reference1_" + self.guid + self.mapa["_2f"],
            ),
            Reference(
                DigestMethod(Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"),
                DigestValue(digestvalue2),
                Id="SignedProperties-Reference_" + self.guid + self.mapa["_20"],
                Type="http://uri.etsi.org/01903#SignedProperties",
                URI="#SignedProperties_" + self.guid + self.mapa["_46"],
            ),
            Id="SignedInfo_" + self.guid + self.mapa["_49"],
        )

        canonicalizedxml = self._c14n(signedinfo, "")
        if self.debug:
            print("*" * 20, "build signedinfo")
            print(canonicalizedxml)
        signature = signproc(canonicalizedxml, "sha256")
        actualdigestencoded = base64.b64encode(signature).decode()
        digestvalue3 = []
        for i in range(0, len(actualdigestencoded), 64):
            digestvalue3.append(actualdigestencoded[i : i + 64])
        digestvalue3 = "\n".join(digestvalue3)

        unsignedproperties = self.unsignedpropertied(
            digestvalue2, tspurl, tspcred, "sha256"
        )

        DOC = Signature(
            signedinfo,
            SignatureValue(
                digestvalue3, Id="SignatureValue_" + self.guid + self.mapa["_5a"]
            ),
            KeyInfo(
                X509Data(X509Certificate(certcontent)),
                Id="KeyInfo_" + self.guid + self.mapa["_2c"],
            ),
            Object(
                QualifyingProperties(
                    signedprop,
                    unsignedproperties,
                    Id="QualifyingProperties_" + self.guid + self.mapa["_4b"],
                    Target="#Signature_" + self.guid + self.mapa["_11"],
                )
            ),
            Id="Signature_" + self.guid + self.mapa["_11"],
        )
        if signedobj is not None:
            DOC.append(signedobj)
        return DOC



#!/usr/bin/env vpython3
# coding: utf-8
import unittest
import os
from subprocess import PIPE, Popen
import sys
from datetime import datetime

from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12
from endesive import plain

import test_cert

tests_root = os.path.dirname(__file__)
fixtures_dir = os.path.join(tests_root, 'fixtures')


def fixture(fname):
    return os.path.join(fixtures_dir, fname)


class PLAINTests(unittest.TestCase):
    def test_plain_signed_attr(self):
        p12 = test_cert.CA().pk12_load(test_cert.cert1_p12, '1234')
        with open(fixture('plain-unsigned.txt'), 'rb') as fh:
            datau = fh.read()
        datas = plain.sign(datau,
            p12[0], p12[1], p12[2],
            'sha256',
            attrs=True
        )
        fname = fixture('plain-signed-attr.txt')
        with open(fname, 'wb') as fh:
            fh.write(datas)

        cmd = [
            'openssl', 'smime', '-verify',
            '-CAfile', test_cert.ca_cert,
            '-content', fixture('plain-unsigned.txt'),
            '-in', fname,
            '-inform', 'der',
        ]
        process = Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()
        assert stderr == b'Verification successful\n'
        assert datau == stdout

    def test_plain_signed_noattr(self):
        p12 = test_cert.CA().pk12_load(test_cert.cert1_p12, '1234')
        with open(fixture('plain-unsigned.txt'), 'rb') as fh:
            datau = fh.read()
        datas = plain.sign(datau,
            p12[0], p12[1], p12[2],
            'sha256',
            attrs=False
        )
        fname = fixture('plain-signed-noattr.txt')
        with open(fname, 'wb') as fh:
            fh.write(datas)

        cmd = [
            'openssl', 'smime', '-verify',
            '-CAfile', test_cert.ca_cert,
            '-content', fixture('plain-unsigned.txt'),
            '-in', fname,
            '-inform', 'der',
        ]
        process = Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()
        assert stderr == b'Verification successful\n'
        assert datau == stdout

    def test_plain_ssl_attr(self):
        cmd = [
            'openssl', 'smime', '-sign',
            '-md', 'sha256',
            '-binary',
            '-CAfile', test_cert.ca_cert,
            '-in', fixture('plain-unsigned.txt'),
            '-out', fixture('plain-ssl-signed-attr.txt'),
            '-outform', 'der',
            '-inkey', test_cert.cert1_key,
            '-passin', 'pass:1234',
            '-signer', test_cert.cert1_cert,
        ]
        process = Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()
        assert b'' == stdout
        assert b'' == stderr

        with open(test_cert.ca_cert, 'rb') as fh:
            trusted_cert_pems = (fh.read(),)
        with open(fixture('plain-unsigned.txt'), 'rb') as fh:
            datau = fh.read()
        with open(fixture('plain-ssl-signed-attr.txt'), 'rb') as fh:
            datas = fh.read()
        (hashok, signatureok, certok) = plain.verify(datas, datau, trusted_cert_pems)
        assert signatureok and hashok and certok

    def test_plain_ssl_noattr(self):
        cmd = [
            'openssl', 'smime', '-sign',
            '-md', 'sha256',
            '-binary', '-noattr',
            '-CAfile', test_cert.ca_cert,
            '-in', fixture('plain-unsigned.txt'),
            '-out', fixture('plain-ssl-signed-noattr.txt'),
            '-outform', 'der',
            '-inkey', test_cert.cert1_key,
            '-passin', 'pass:1234',
            '-signer', test_cert.cert1_cert,
        ]
        process = Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()
        assert b'' == stdout
        assert b'' == stderr

        with open(test_cert.ca_cert, 'rb') as fh:
            trusted_cert_pems = (fh.read(),)
        with open(fixture('plain-unsigned.txt'), 'rb') as fh:
            datau = fh.read()
        with open(fixture('plain-ssl-signed-noattr.txt'), 'rb') as fh:
            datas = fh.read()
        (hashok, signatureok, certok) = plain.verify(datas, datau, trusted_cert_pems)
        assert signatureok and hashok and certok



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12
from endesive import plain


def main():
    with open('ca/demo2_user1.p12', 'rb') as fp:
        p12 = pkcs12.load_key_and_certificates(fp.read(), b'1234', backends.default_backend())
    datau = open('plain-unsigned.txt', 'rb').read()
    datas = plain.sign(datau,
        p12[0], p12[1], p12[2],
        'sha512',
        attrs=True,
        pss=True
    )
    open('plain-signed-pss.txt', 'wb').write(datas)


main()
